/* 
 *  mod_cachem.c -- Apache2 DSO sahred memory manager module
 * Byron Young
 * 2005
 */

/**
 *@file mod_cachem.c
 *@brief mod_cachem source code
 *@author Byron Young
 *@version 0.1
 *@bug report to spamiccling@yahoo.com
 *@bug report to bkyoung@users.sourceforge.net
 */

#include <httpd.h>
#include <http_config.h>
#include <http_core.h>
#include <http_main.h>
#include <http_connection.h>
#include <http_request.h>
#include <http_log.h>
#include <ap_mmn.h>
#include <apr_dbm.h>
#include <mpm.h>
#include <apr.h> /* apr_uint32_t */
#include <apr_strings.h>
#include <apr_tables.h>
#include <apr_lib.h>
#include <apr_file_io.h>
#include <apr_file_info.h>
#include <apr_pools.h>
#include <http_protocol.h>
#include <ap_config.h>
#include <mod_status.h>
#include <apr_strings.h>
#include <util_script.h>
#include <util_filter.h>
#include <apr_version.h>
#if APR_HAS_THREADS
#include <apr_thread_mutex.h>
#endif /* APR_HAS_THREADS */
#include <ap_mpm.h>
#include <apr_proc_mutex.h>
#include <apr_global_mutex.h>
#include <apr_network_io.h>
#include "apr_fnmatch.h"
#include <apr_rmm.h>
#include <apr_shm.h>
#include <apr_optional.h>
#define APR_WANT_STRINGS
#define APR_WANT_MEMFUNC
#include <apr_want.h>

#if !defined(OS2) && !defined(WIN32) && !defined(BEOS) && !defined(NETWARE)
#include <unixd.h>
#define MOD_COUNTM_SET_MUTEX_PERMS
#endif
#include <ctype.h>
#include <unistd.h>
#include <stdarg.h>

#include "mod_cachem_private.h"
#include "mod_cachem.h"

module AP_MODULE_DECLARE_DATA cachem_module;
command_rec cachem_cmds[];

/**
 *@fn cachem_id(const char * s, apr_pool_t * p)
 *@param s A unique string, common to a specific module, but uniqe among all modules.
 *@param p A memory pool.
 *@brief returns a unique, to all modules, id number.
 *@return a unique number.
 *@return CACHEM_INVALID_ID a DBM error.
 *If s is NULL, then returns a generic id.
 *s may contain only apr_isalnum '.' '_' '-' characters.
 *s may not exceed 255 characters.
 *If s is already within DBM, then the stored ID is returned.
 *If s is not present within DBM, it is inserted.
 */
cachem_id_t
cachem_id(const char * s, apr_pool_t * p) {
  apr_status_t rs;
  p_cachem_header_t pch=(p_cachem_header_t) cachem.pData;
  cachem_id_t id;
  int i=0;
  int len;
  apr_dbm_t * dbm=NULL;
  apr_datum_t key, data;

  if (s) {
    if ((len=strlen(s)) > 255) return CACHEM_INVALID_ID;
    while ((s[i]!='\0') && ( (apr_isalnum(s[i])) || (s[i]=='_') || (s[i]=='-') || (s[i]=='.')) ) ++i;
    if (len!=i) return CACHEM_INVALID_ID;

    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));
    key.dptr=(char *) s;
    key.dsize=strlen(s)+1;
    rs=apr_dbm_open(&dbm, cachem.dbmname, APR_DBM_RWCREATE,
		    APR_OS_DEFAULT, p);
    if (rs!=APR_SUCCESS) goto err;
    if(apr_dbm_exists(dbm, key)) {
      rs=apr_dbm_fetch(dbm, key, &data);
      id = (cachem_id_t) atol(data.dptr);
      apr_dbm_freedatum(dbm, data);
      apr_dbm_close(dbm);
      return id;
    } else {
      ++pch->u_id;
      id = pch->u_id;
      data.dptr=apr_psprintf(p, "%u", id);
      data.dsize=strlen(data.dptr) + 1;
      rs=apr_dbm_store(dbm, key, data);
      if (rs!=APR_SUCCESS) goto err;
      apr_dbm_close(dbm);
      return id;
    }
  } else {
    ++pch->u_id;
    id = pch->u_id;
    return id;
  }
 err:
  if (dbm!=NULL) apr_dbm_close(dbm);
  ap_log_perror(APLOG_MARK, APLOG_CRIT, rs, p,
		"Opening CACHEMDBM");
  return CACHEM_INVALID_ID;
} /* cachem_id */

/**
 *@fn cachem_shm_destory(server_rec *s, apr_pool_t *p)
 *@param s The server record.
 *@param p THe pool.
 *@brief destroy the shared memory, either named or anonymous.
 *only call from non detached process, so lock not required.
 */
static void
cachem_shm_destory(server_rec *s, apr_pool_t *p) {
  p_cachem_config_t p_cc=NULL;
  apr_status_t rs=APR_SUCCESS;

  /* get base server configuration */
  p_cc = (p_cachem_config_t) ap_get_module_config(s->module_config,
						 &cachem_module);

  if (p_cc->sza_shmfile) {
    /* named shared memory */
    if (cachem.shm == NULL) {
      /* non detached process did not access shared memory. */
      /* attempt to attach to a existing shm, */
      /* possibly created by other detached process. */
      rs=apr_shm_attach(&cachem.shm, p_cc->sza_shmfile, p);
      if (rs!=APR_SUCCESS) {
	/* could not attach to shared memory, must not exist. */
	return;
      }
    }
    
    /* able to attach to named shared memory OR was already attached. */
    /* destory it. */
    rs = apr_shm_destroy(cachem.shm);
    cachem.shm=NULL;
    return;
  } else {
    /* anon shared memory */
    if (cachem.shm == NULL) {
      /* never created */
      return;
    } else {
      rs = apr_shm_destroy(cachem.shm);
      cachem.shm=NULL;
      return;
    }    
  }
} /* cachem_shm_destory */

/**
 *@fn cachem_named_shm_init(request_rec *r)
 *@param r The request record
 *@brief initialize named (as opposed to anon shared memory for a process.)
 *shared memory lasts until restart (as opposed to exit)
 *first accessor will initialize it, others will attach.
 */
static int
cachem_named_shm_init(request_rec *r) {
  p_cachem_config_t p_cc=NULL;
  apr_status_t rs=APR_SUCCESS;
  apr_pool_t * p = r->pool;
  apr_pool_t * pconf = r->server->process->pconf;
  int ret_val=0;

  /* get base server configuration */
  p_cc = (p_cachem_config_t) ap_get_module_config(r->server->module_config,
						 &cachem_module);

  rs=apr_global_mutex_lock(cachem_mutex);
  if (cachem.shm) {
    goto release_lock;
  }
  
  /* attempt to attach to shared memory area. */
  rs=apr_shm_attach(&cachem.shm, p_cc->sza_shmfile, pconf);
  if (rs==APR_SUCCESS) {
    /* already created, just attached to it. */
    cachem.pData = apr_shm_baseaddr_get(cachem.shm);
    cachem.nSize = apr_shm_size_get(cachem.shm);
  } else {
    /* attempt to create shared memory area. */
    cachem.shm=NULL;
    if ((rs = apr_shm_create(&cachem.shm, 
			     p_cc->shmsize, 
			     p_cc->sza_shmfile,
			     pconf)) != APR_SUCCESS) {
      if (rs==APR_ENOTIMPL) {
	ap_log_perror(APLOG_MARK, APLOG_ERR, rs, p,
		     "Named shared memory is not supported");
      } else {
	ap_log_perror(APLOG_MARK, APLOG_ERR, rs, p,
		     "Cannot allocate shared memory:");
      }
      ret_val=1;
      goto release_lock;
    }
    cachem.pData = apr_shm_baseaddr_get(cachem.shm);
    cachem.nSize = apr_shm_size_get(cachem.shm);
    cachem_init();
  }

 release_lock:
  rs = apr_global_mutex_unlock(cachem_mutex);
  return ret_val;
} /* cachem_named_shm_init */

/**
 *@fn cachem_shm_detach()
 **brief detach a process from (named/anon) shared memory
 *call during detached process termination.
 */
static void
cachem_shm_detach() {
  apr_status_t rs;

  if (cachem.shm) {
    rs=apr_shm_detach(cachem.shm);
    cachem.shm=NULL;
  }
  return;
} /* cachem_shm_detach */

/**
 *@fn cachem_init()
 *@brief Initialize the session cache.
 *@return void
 *If called during non detached process, no lock required.
 *If called during detached process, lock required.
 */
static void
cachem_init() {
  p_cachem_header_t pch=(p_cachem_header_t) cachem.pData;

  memset(cachem.pData, 0, cachem.nSize);
  pch->o_st=CACHEM_DATA_START;
  pch->o_fr=CACHEM_DATA_START;
  pch->u_size=(unsigned int) cachem.nSize;
  pch->o_bptop=pch->u_size;
  return;
} /* cachem_init */

/**
 *@fn cachem_add(void * data, unsigned int dsize, cachem_id_t id)
 *@brief Adds record to cache.
 *The index is the first unused one, or it is allocated.
 *The block is inserted if using an unused index, or allocated from free.
 *Memory may be moved, pointers aquired via cachem_get may become invalid.
 */
m_index_t
cachem_add(void * data, unsigned int dsize, cachem_id_t id) {
  m_index_t index=0; /* The bptr to use. Either an unused one, or allocated. */
  p_cachem_header_t pch=(p_cachem_header_t) cachem.pData;
  p_cachem_bptr_t pbptr;
  p_cachem_bptr_t prevbptr;
  unsigned int bsize=APR_ALIGN(dsize, 4); /* block size */
  unsigned int tsize=bsize + sizeof(cachem_bptr_t); /* total free bytes. */
  void * ptr;
  void * ptr_s;
  void * ptr_d;
  unsigned int u_bmove;
  unsigned int o;
  unsigned int o_prev;
  unsigned int o_beg;
  int bAlloc=1;

  /* attempt to find the first unused block pointer. */
   for (o=CACHEM_FIRST_BP_OFFSET(pch); o>=pch->o_bptop;
       o-=sizeof(cachem_bptr_t), ++index){
     //    pbptr=(p_cachem_bptr_t) (cachem.pData + o);
    pbptr=(p_cachem_bptr_t) ((char *)cachem.pData + o);
    if (pbptr->o_beg==0) {
      /* found an unused one */
      --bAlloc;
      tsize-=sizeof(cachem_bptr_t);
      break;
    }
  }

  if (tsize > (pch->o_bptop - pch->o_fr) ) {
    /* overflow */
    return CACHEM_INVALID_INDEX;
  }

  if (bAlloc) {
    /* allocate block pointer, add to free */
    pch->o_bptop-=sizeof(cachem_bptr_t); /* o */
    pbptr = (p_cachem_bptr_t) (((char *) cachem.pData) + pch->o_bptop);
    ptr = ((char *) cachem.pData) + pch->o_fr;
    pbptr->o_beg=pch->o_fr;
    pbptr->u_size=dsize;
    pbptr->id=id;
    pbptr->pid=getpid();
    pch->o_fr+=bsize;
    ++pch->u_blks;
    memset(ptr, 0, bsize);
    memmove(ptr, data, dsize);
    return index;
  }

  if (o==CACHEM_FIRST_BP_OFFSET(pch)) {
    /* inserting the first block pointer. */
    /* beginning of area to move is the top of the cache */
    o_beg=pch->o_st;
  } else {
    /* previous bptr is just prior to this one. */
    o_prev=o + sizeof(cachem_bptr_t);
    prevbptr = (void *) (((char *) cachem.pData) + o_prev);
    /* beginning of area to move is just after prev block */
    o_beg=prevbptr->o_beg + APR_ALIGN(prevbptr->u_size, 4);
  }
  /* source */
  ptr_s = ((char *) cachem.pData) + o_beg;
  /* size of area to move is */
  u_bmove = pch->o_fr - o_beg;
  /* destination */
  ptr_d = (void *) (((char *)ptr_s) + bsize);

  memmove(ptr_d, ptr_s, u_bmove);
  memset(ptr_s, 0, bsize);
  memmove(ptr_s, data, dsize);
  pbptr->o_beg=o_beg;
  pbptr->u_size=dsize;
  pbptr->id=id;
  pbptr->pid=getpid();
  pch->o_fr+=bsize;

  /* adjust all used bptrs after */
  for (o-=sizeof(cachem_bptr_t);
       o>=pch->o_bptop;
       o-=sizeof(cachem_bptr_t)){
    pbptr=(p_cachem_bptr_t) (((char *) cachem.pData) + o);
    if (pbptr->o_beg!=0) {
      pbptr->o_beg+=bsize;
    }
  }

  return index;
} /* cachem_add */

/**
 *@fn cachem_delete(m_index_t index)
 *@param index The index of the record to delete.
 *@brief Deletes the block from the cache.
 *The index slot is simply zeroed out.
 *If the deleted block empties the cache, info is adjusted.
 *Memory may be moved. Pointers obtained with cachem_get may be invalidated.
 *Only an empty cache "ends" with an unused slot.
 *@return The index of the block deleted.
 *@return CACHEM_INVALID_INDEX Invalid index given.
 */
m_index_t
cachem_delete(m_index_t index) {
  p_cachem_header_t pch=(p_cachem_header_t) cachem.pData;
  p_cachem_bptr_t pbptr;
  p_cachem_bptr_t pbptr_a;
  unsigned int io;
  void * ptr;
  void * ptr_s;
  void * ptr_d;
  size_t sz;

  if (pch->u_blks==0) {
    /* invalid index */
    index=CACHEM_INVALID_INDEX;
    goto invalid_index;
  }

  io = pch->u_size - (index + 1) * sizeof(cachem_bptr_t);
  if (io < pch->o_bptop) {
    /* invalid index. */
    index=CACHEM_INVALID_INDEX;
    goto invalid_index;
  }

  pbptr= (p_cachem_bptr_t) (((char *) cachem.pData) + io);
  if (pbptr->o_beg==0) {
    /* already deleted */
    index=CACHEM_INVALID_INDEX;
    goto invalid_index;
  }

  /* is this the last one? */
  if (io == pch->o_bptop) {
    /* last one */
    pch->o_bptop+=sizeof(cachem_bptr_t);
    io+=sizeof(cachem_bptr_t);
    --pch->u_blks;
    pch->o_fr-=APR_ALIGN(pbptr->u_size, 4);
    ptr = ((char *) cachem.pData) + pbptr->o_beg;
    memset(ptr, 0, APR_ALIGN(pbptr->u_size, 4));
    memset(pbptr, 0, sizeof(cachem_bptr_t));

    /* work down, to next used block pointer */
    for(;
	io < pch->u_size;
	io+=sizeof(cachem_bptr_t), pch->o_bptop+=sizeof(cachem_bptr_t)) {
      pbptr = (p_cachem_bptr_t) (((char *) cachem.pData) + io);
      if (pbptr->o_beg!=0) {
	break;
      }
    }

    return index;
  }
  

  /* must shrink cache */
  /* source */
  ptr_s = ((char *) cachem.pData) + pbptr->o_beg + APR_ALIGN(pbptr->u_size, 4);
  /* destination */
  ptr_d = ((char *) cachem.pData) + pbptr->o_beg;
  /* size */
  sz = (size_t) (pch->o_fr - (pbptr->o_beg + APR_ALIGN(pbptr->u_size, 4)));
  memmove(ptr_d, ptr_s, sz);
  --pch->u_blks;
  pch->o_fr-=APR_ALIGN(pbptr->u_size, 4);
  memset(((char *) cachem.pData) + pch->o_fr, 0, APR_ALIGN(pbptr->u_size, 4));

  /* adjust remaining block pointers */
  for (io=pch->u_size - (index + 2) * sizeof(cachem_bptr_t);
       io>=pch->o_bptop;io-=sizeof(cachem_bptr_t)) {
    pbptr_a= (p_cachem_bptr_t) (((char *) cachem.pData) + io);
    if (pbptr_a->o_beg!=0) {
#if defined (DEBUG)
      void * pt;

      pt = ((char *) cachem.pData) + pbptr_a->o_beg;
#endif
      pbptr_a->o_beg-=APR_ALIGN(pbptr->u_size, 4);
#if defined (DEBUG)
      pt = ((char *) cachem.pData) + pbptr_a->o_beg;
#endif
    }
  }

  /* zero out block pointer */
  memset(pbptr, 0, sizeof(cachem_bptr_t));

 invalid_index:
  return index;
} /* cachem_delete */

/**
 *@param id The module id to match. (may be CACHEM_INVALID_ID)
 *@brief Delete all records matching current process pid and id from cache.
 *@return void
 */
void
cachem_delete_pid(cachem_id_t id) {
  m_index_t index;

  while( (index=cachem_get_pid(id, CACHEM_INVALID_INDEX))!=CACHEM_INVALID_INDEX) {
    index=cachem_delete(index);
  }
  return;
} /* cachem_delete_pid */

/**
 *@fn cachem_next(m_index_t index)
 *@param index Return index after index.
 *@brief returns the next index, after given index.
 *If index==CACHEM_INVALID_INDEX, then returns the first index.
 *@return index number.
 *@return if returned index==CACHEM_INVALID_INDEX, then no more entries.
 */
static m_index_t
cachem_next(m_index_t index) {
  p_cachem_header_t pch=(p_cachem_header_t) cachem.pData;
  unsigned int o;
  p_cachem_bptr_t pbptr;
  unsigned int cindex=0;

  if (pch->u_blks==0) {
    cindex = CACHEM_INVALID_INDEX;
    goto return_cindex;
  }

  for (o=CACHEM_FIRST_BP_OFFSET(pch); o>=pch->o_bptop;
       o-=sizeof(cachem_bptr_t), ++cindex){
    pbptr=(p_cachem_bptr_t) (((char *) cachem.pData) + o);
    if (pbptr->o_beg==0) continue;
    if (index==CACHEM_INVALID_INDEX) goto return_cindex;
    if (cindex>index) goto return_cindex;
  }
  cindex = CACHEM_INVALID_INDEX;
 return_cindex:
  return cindex;
} /* cachem_next */

/**
 *@fn cachem_get(m_index_t index, unsigned int * dsize)
 *@param index The index to retrieve
 *@param dsize OUT:The size of the record in the block.
 *@brief aquire a pointer to the block, in the cache, at index.
 *returned pointer is supposed to be 4 aligned.
 *@return pointer to block.
 *@return NULL invalid index given.
 *Aquire and use pointer within cachem_mutex.
 */
void *
cachem_get(m_index_t index, unsigned int * dsize) {
  p_cachem_header_t pch=(p_cachem_header_t) cachem.pData;
  p_cachem_bptr_t pbptr;
  void * pblk;
  unsigned int o;

  o=pch->u_size - (index + 1) * sizeof(cachem_bptr_t);
  if (o<pch->o_bptop) return NULL;

  pbptr = (p_cachem_bptr_t) (((char *) cachem.pData) + o);
  if (pbptr->o_beg==0) return NULL;
  pblk =  ((char *) cachem.pData) + pbptr->o_beg;
  *dsize=pbptr->u_size;
  return pblk;
} /* cachem_get */

/**
 *@fn cachem_get_pid(cachem_id_t id, m_index_t index)
 *@param id The module id to match.
 *@param index The start index.
 *@return CACHEM_INVALID_INDEX no records were found.
 *@return index
 *@return index==CACHEM_INVALID_INDEX
 *brief returns the index of block.
 *The returned index will match:
 * With pid of the current process.
 * With id given. (May be CACHEM_INVALID_ID)
 * After index. (If CACHEM_INVALID_INDEX, then first.)
 *Call function until returns CACHEM_INVALID_INDEX to find all records.
 *If id==INVALID, then matches invalid id's (at this time just matches)
 *If index==INVALID, the returns first matching.
 *If index then returns the first after index matching pid and id.
 */
m_index_t
cachem_get_pid(cachem_id_t id, m_index_t index) {
  p_cachem_header_t pch=(p_cachem_header_t) cachem.pData;
  unsigned int o;
  p_cachem_bptr_t pbptr;
  unsigned int cindex=0;
  unsigned int pid=getpid();

  if (pch->u_blks==0)
    return CACHEM_INVALID_INDEX;

  for (o=CACHEM_FIRST_BP_OFFSET(pch); o>=pch->o_bptop;
       o-=sizeof(cachem_bptr_t), ++cindex){
    pbptr=(p_cachem_bptr_t) (((char *) cachem.pData) + o);
    if (pbptr->o_beg==0) continue;
    if (pbptr->pid!=pid) continue;
    if (pbptr->id!=id) continue; /* TODO: match all invalid id's.*/
    /* matched pid and id */
    if ( (index==CACHEM_INVALID_INDEX) || (cindex > index)) {
      return cindex;
    }
  }
  return CACHEM_INVALID_INDEX;
} /* cachem_get_pid */

/**
 *@fn cachem_get_id(cachem_id_t id, m_index_t index)
 *@param index The start index.
 *@param bLock Use the lock while getting.
 *@return The first index after given index, with id.
 *If index is CACHEM_INVALID_INDEX, then returns the first index.
 *An id==CACHEM_INVALID_ID is an error.
 *Matches, ignoring pid.
 */
m_index_t
cachem_get_id(cachem_id_t id, m_index_t index) {
  p_cachem_header_t pch=(p_cachem_header_t) cachem.pData;
  unsigned int o;
  p_cachem_bptr_t pbptr;
  unsigned int cindex=0;

  if (id==CACHEM_INVALID_ID)
    return CACHEM_INVALID_INDEX;

  if (pch->u_blks==0) {
    cindex = CACHEM_INVALID_INDEX;
    goto return_cindex;
  }

  for (o=CACHEM_FIRST_BP_OFFSET(pch); o>=pch->o_bptop;
       o-=sizeof(cachem_bptr_t), ++cindex){
    pbptr=(p_cachem_bptr_t) (((char *) cachem.pData) + o);
    if (pbptr->o_beg==0) continue;
    if (pbptr->id==id) {
      if (index==CACHEM_INVALID_INDEX) goto return_cindex;
      if (cindex > index) goto return_cindex;
    }
  }
  cindex = CACHEM_INVALID_INDEX;
 return_cindex:
  return cindex;
} /* cachem_get_id */

/**
 *@fn cachem_info(p_cachem_stats_t pst)
 *@bparam pst Allocated structure for return information.
 *@brief returns information.
 *
 */
void
cachem_info(p_cachem_stats_t pst) {
  p_cachem_header_t pch=(p_cachem_header_t) cachem.pData;

  pst->u_blks = pch->u_blks;
  pst->free = pch->o_bptop - pch->o_fr;  
  pst->size=pch->u_size;
  return;
} /* cachem_info */

/**
 *@fn cachem_lock()
 *@return the result
 *@brief locks the cachem_mutex
 */
apr_status_t
cachem_lock() {
  return apr_global_mutex_lock(cachem_mutex);
} /* cachem_lock */

/**
 *@fn cachem_unlock()
 *@return the result
 *@brief unlocks the cachem_mutex
 */
apr_status_t
cachem_unlock() {
  return apr_global_mutex_unlock(cachem_mutex);
} /* cachem_unlock */

/**
 *@fn cachem_handler_init(request_rec * r)
 *@param r The request record.
 *@brief Call within each modules method_handler.
 *@return 0=success 1=error, do not continue.
 */
int
cachem_handler_init(request_rec * r) {

  if (cachem.shm==NULL) {
    return cachem_named_shm_init(r);
  }
  return 0;
} /* cachem_handler_init */

static apr_status_t
cachem_post_config_cleanup(void * data) {
  apr_status_t rs;
  p_cachem_cleanup_data_t p_ccd = (p_cachem_cleanup_data_t) data;
  apr_pool_t * p =p_ccd->p;
  /*  int bDetached=p_ccd->bDetached; */
  server_rec * s=p_ccd->s;

  cachem_shm_destory(s,p);

  /* release global mutex */
  if ((rs=apr_global_mutex_destroy(cachem_mutex))!=APR_SUCCESS) {
    ap_log_error(APLOG_MARK, APLOG_CRIT, rs, NULL, 
		 "Destroy global mutex failed.");
  }

  return APR_SUCCESS;
} /* cachem_post_config_cleanup */

static int
cachem_hook_post_config(apr_pool_t *p, apr_pool_t * plog,
			apr_pool_t *ptemp, server_rec *s) {
  p_cachem_config_t p_cc=NULL;
  p_cachem_cleanup_data_t p_ccd;
  apr_status_t rs;
  const char * dbm1, * dbm2;

  /* get the global server cachem_config. */
  p_cc = ap_get_module_config(s->module_config, &cachem_module);

  if (p_cc->sza_lockfile==NULL) {
    return DECLINED;
  }

  if (p_cc->sza_dbmname==NULL) {
    return DECLINED;
  }

  memset(&cachem, 0, sizeof(cachem));
  apr_dbm_get_usednames(p, p_cc->sza_dbmname, &dbm1, &dbm2);
  if (dbm1) apr_file_remove(dbm1, p);
  if (dbm2) apr_file_remove(dbm2, p);
  cachem.dbmname=apr_pstrdup(p, p_cc->sza_dbmname);

  if (!p_cc->sza_shmfile) {
    /* using ANON shared memory */
    rs = apr_shm_create(&cachem.shm, 
			p_cc->shmsize, 
			NULL,
			p);
    if (rs==APR_ENOTIMPL) {
      ap_log_perror(APLOG_MARK, APLOG_ERR, rs, p,
		    "ANON shared memory is not available:");
      return DECLINED;
    } else if (rs!=APR_SUCCESS) {
      ap_log_perror(APLOG_MARK, APLOG_ERR, rs, p,
		    "Cannot allocate shared memory:");
      return DECLINED;
    }
    cachem.pData = apr_shm_baseaddr_get(cachem.shm);
    cachem.nSize = apr_shm_size_get(cachem.shm);
    cachem_init();
  }

  /* create a global mutex */
  rs=apr_global_mutex_create(&cachem_mutex, p_cc->sza_lockfile,
			     APR_LOCK_DEFAULT, p);
    if (rs!=APR_SUCCESS) {
      ap_log_perror(APLOG_MARK, APLOG_CRIT, rs, p,
		    "%s: Create global mutex failed", p_cc->sza_lockfile);
    return DECLINED;
  }
  /* and adjust permissions, if necessary */
#ifdef MOD_CACHEM_SET_MUTEX_PERMS
  rs=unixd_set_global_mutex_perms(cachem_mutex);
  if (rs != APR_SUCCESS) {
    ap_log_error(APLOG_MARK, APLOG_CRIT, ret, NULL, 
		 "Cannot adjust process mutex permissions.");
    apr_global_mutex_destory(cachem_mutex);
    return DECLINED;
  }
#endif /* MOD_CACHEM_SET_MUTEX_PERMS */

  p_ccd=apr_pcalloc(p, sizeof(*p_ccd));
  p_ccd->p=p;
  p_ccd->bDetached=0;
  p_ccd->s=s;
  apr_pool_cleanup_register(p, (void *) p_ccd,
			    cachem_post_config_cleanup,
			    apr_pool_cleanup_null);
  
  return OK;
} /* cachem_hook_post_config */

static apr_status_t
cachem_child_init_cleanup(void * parm) {
  /*  p_cachem_cleanup_data_t p_ccd = (p_cachem_cleanup_data_t) parm;
      int bDetached=p_ccd->bDetached;
      apr_pool_t * p =p_ccd->p;
      server_rec * s = p_ccd->s; 
      apr_status_t rs; */
  //  m_index_t index=CACHEM_INVALID_INDEX;

  /* delete all from cache with this pid with id==CACHEM_INVALID_ID */
  cachem_lock();
  cachem_delete_pid(CACHEM_INVALID_ID);
  cachem_unlock();
  cachem_shm_detach();
  return APR_SUCCESS;
} /* cachem_child_init_cleanup */

static void
cachem_hook_child_init(apr_pool_t *p, server_rec *s) {
  p_cachem_config_t p_cc=NULL;
  p_cachem_cleanup_data_t p_ccd;
  apr_status_t rs;

  p_cc = ap_get_module_config(s->module_config, &cachem_module);

  /* bring global mutex into child process */
  rs = apr_global_mutex_child_init(&cachem_mutex, p_cc->sza_lockfile, p);
  if (rs) {
    ap_log_error(APLOG_MARK, APLOG_CRIT, rs, NULL, 
		  "Error initializing global mutex for pid=%d", getpid());
  }

  if (p_cc->sza_shmfile) {
    /* initialize values (named shared memory) */
    cachem.shm=NULL;
    cachem.pData = NULL;
    cachem.nSize = 0;
  } else {
    /* inherit values (anon shared memory) */
    /* pSessionCacheDataMM */
    /* pSessionCacheData */
    /* nSessionCacheSize */
    cachem.pData = apr_shm_baseaddr_get(cachem.shm);
    cachem.nSize = apr_shm_size_get(cachem.shm);
  }

  p_ccd=apr_pcalloc(p, sizeof(*p_ccd));
  p_ccd->p=p;
  p_ccd->bDetached=1;
  p_ccd->s=s;
  apr_pool_cleanup_register(p, (void *) p_ccd,
			    cachem_child_init_cleanup,
			    apr_pool_cleanup_null);
  return;
} /* cachem_hook_child_init */

#ifdef DEBUG
/**
 *@param s A numeric string.
 *@brief Increments the string by one.
 */
static void
cachem_dbg_str_inc(char * s) {
  int i;

  for(i=strlen(s)-1; i>=0; --i) {
    ++s[i];
    if (s[i]>'9') {
      s[i]='0';
    } else {
      break;
    }
  }
  return;
} /* cachem_dbg_str_inc */

/**
 *@brief Do some debugging.
 */
static m_index_t
cachem_dbg(cachem_id_t id, request_rec * r) {
  char buf[128];
  char tbuf[64];
  char nstr[]="0000";
  m_index_t index=!CACHEM_INVALID_INDEX;
  m_index_t last_index;
  apr_pool_t * p=r->pool;

  /* create some strings */
  memset(tbuf, 0, sizeof(tbuf));
  apr_ctime(tbuf, r->request_time);

  /* fill up cache to overflow */
  for(index=!CACHEM_INVALID_INDEX;
	index!=CACHEM_INVALID_INDEX;
	cachem_dbg_str_inc(nstr)) {
    strcpy(buf, tbuf);
    strcat(buf, " ");
    strcat(buf, nstr);
    last_index=index;
    index=cachem_add(buf, strlen(buf)+1, id);
  }

  // delete all from cache
  //  index=CACHEM_INVALID_INDEX;
  //  while ( CACHEM_INVALID_INDEX != (index=cachem_next(index)) ) {
  //    cachem_delete(index);
  //  }

  // delete the last one
  //  index=last_index;
  //  cachem_delete(index);

  // delete the last - 2 and last - 1 
  //cachem_delete(index-2);
  //cachem_delete(index-1);

  // delete the first one
  cachem_delete(0);
  // delete the third one
  cachem_delete(2);
  // add in a new string (should be index 0)
  strcpy(buf, tbuf);
  strcat(buf, " ");
  strcat(buf, "AAAA");
  index=cachem_add(buf, strlen(buf)+1, id);
  // add in a new string (should be index 2)
  strcpy(buf, tbuf);
  strcat(buf, " ");
  strcat(buf, "BBBB");
  index=cachem_add(buf, strlen(buf)+1, id);
  return 0;
} /* cachem_dbg */

/* The sample content handlers */
static int
cachem_handler(request_rec *r) {
  cachem_id_t id;

  if (strcmp(r->handler, "cachem")) {
    return DECLINED;
  }

  cachem_handler_init(r);
  id = cachem_id(__FILE__, r->pool);
  cachem_dbg(id, r);
  r->content_type = "text/html";      
  
  if (!r->header_only)
    ap_rputs("The sample page from mod_cachem.c\n", r);
  return OK;
}
#endif /* DEBUG */

/*
 *@fn cachem_set_shmfile(cmd_parms *cmd, void *dummy, const char *arg1, const char * arg2)
 *@brief Sets the  p_cc->sza_shmfile string from CACHEM value given in configuration file.
 *Called during the per-server-configuration if specified in the config file.
 *@param cmd The command data structure.
 *@param dummy Unused, since no per-directory-configuration possible.
 *@param arg1 The value specified in the configuration script.
 *@param arg2 The value specified in the configuration script.
 *@return NULL
 *@return char * string to display as error.
 */
static const char *
cachem_set_shmfile(cmd_parms *parms, void *dummy, const char *arg1, const char * arg2) {
  long val;
  unsigned int ui_val;
  apr_size_t size;
  const char * err;

  if (err=ap_check_cmd_context(parms, GLOBAL_ONLY)) return err;

  /* the server being processed server searchm_config info. */
  p_cachem_config_t p_cc= ( p_cachem_config_t)
    ap_get_module_config(parms->server->module_config, &cachem_module);

  if (arg1 && (*arg1!='\0') && (strcmp(arg1, "NULL")!=0)) {
    p_cc->sza_shmfile = apr_pstrdup(parms->pool, arg1);
    p_cc->bSet_shmfile=1;
    apr_file_remove(arg1, parms->pool);
  }

  /* some max/min value check may be necessay */
  if (arg2 && (*arg2!='\0') && (strcmp(arg2, "NULL")!=0) ) {
    if ( ((val=atol(arg2)) % 8)==0) {
      return "CACHEM size value must be mod 8";
    }
    ui_val = (unsigned int) val;
    size=(apr_size_t) ui_val;
    p_cc->shmsize= size;
    p_cc->bSet_shmsize=1;
  }

  return NULL;  
} /* cachem_set_shmfile */

/**
 *@fn cachem_set_lockfile(cmd_parms *parms, void *dummy, const char *arg)
 *@brief Sets the  p_cc->sza_lockfile string from CACHEMlockFile value given
 *in configuration file.
 *Called during the per-server-configuration if specified in the config file.
 *@param parms The command data structure.
 *@param dummy Unused, since no per-directory-configuration possible.
 *@param arg The value specified in the configuration script.
 *@return NULL
 */
static const char *
cachem_set_lockfile(cmd_parms *parms, void *dummy, const char *arg) {
  /* the server being processed server cachem_config info. */
  p_cachem_config_t p_cc= ( p_cachem_config_t )
    ap_get_module_config(parms->server->module_config, &cachem_module);
  char * filepath;

  filepath = ap_server_root_relative(parms->pool, arg);
  if (!filepath) {
    return "Invalid CACHEMlockfile";
  }

  p_cc->sza_lockfile = (char *) apr_pstrdup(parms->pool, arg);
  p_cc->bSet_lockfile=1;
  return NULL;
} /* cachem_set_lockfile */

/**
 *@fn cachem_set_dbmname(cmd_parms *parms, void *dummy, const char *arg)
 *@brief Sets the  p_cc->sza_dbmname string from CACHEMDBM value given
 *in configuration file.
 *Called during the per-server-configuration if specified in the config file.
 *@param parms The command data structure.
 *@param dummy Unused, since no per-directory-configuration possible.
 *@param arg The value specified in the configuration script.
 *@return NULL
 */
static const char *
cachem_set_dbmname(cmd_parms *parms, void *dummy, const char *arg) {
  /* the server being processed server cachem_config info. */
  p_cachem_config_t p_cc= ( p_cachem_config_t )
    ap_get_module_config(parms->server->module_config, &cachem_module);
  char * filepath;

  filepath = ap_server_root_relative(parms->pool, arg);
  if (!filepath) {
    return "Invalid CACHEMDBM";
  }

  p_cc->sza_dbmname = (char *) apr_pstrdup(parms->pool, arg);
  p_cc->bSet_dbmname=1;
  return NULL;
} /* cachem_set_dbmname */

/**
 *@fn cachem_per_server_config(apr_pool_t * p, server_rec *s)
 *@param p The config pool.
 *@param s The server record.
 *@brief Server startup configuration routine.
 *@return Pointer to the module_config record.
 */
static void *
cachem_per_server_config(apr_pool_t *p, server_rec *s) {
  cachem_config_t def_sc=INITIALIZE_CACHEM_CONFIG;
  p_cachem_config_t p_cc=NULL;

  p_cc = (p_cachem_config_t) apr_pcalloc(p, sizeof(cachem_config_t));

  memcpy(p_cc, &def_sc, sizeof(def_sc));
  return (void *) p_cc;
} /* cachem_per_server_config */

static int
cachem_ext_status_hook(request_rec *r, int flags) {
  p_cachem_config_t p_cc;
  cachem_stats_t cs;

  /* get base server configuration */
  p_cc = (p_cachem_config_t) ap_get_module_config(r->server->module_config,
						  &cachem_module);

  if (flags & AP_STATUS_SHORT)
    return OK;

  ap_rputs("<hr>\n", r);
  ap_rputs("<p>mod_cachem loaded.</p>\n", r);

  cachem_handler_init(r);
  cachem_info(&cs);
  ap_rprintf(r, "Blocks:%u Total:%u Used:%u Free:%u", cs.u_blks, cs.size, cs.size - cs.free, cs.free);
#ifdef DEBUG
 {
  cachem_id_t id;
  m_index_t index=CACHEM_INVALID_INDEX;
  char * pstr;
  unsigned int dsize;

  cachem_lock();
  id = cachem_id(__FILE__, r->pool);
  while((index=cachem_get_id(id, index))!=CACHEM_INVALID_INDEX) {
    pstr=(char *) cachem_get(index, &dsize);
    ap_rputs("<p>", r);
    ap_rputs(pstr, r);
    ap_rputs("</p>", r);
  }
  //  while((index=cachem_get_id(id, CACHEM_INVALID_ID))!=CACHEM_INVALID_INDEX) {
  //    index=cachem_delete(index);
  //  }
  cachem_unlock();
 }
#endif /* DEBUG */
  return OK;
} /* cachem_ext_status_hook */

static int
cachem_hook_pre_config(apr_pool_t *pconf, apr_pool_t *plog, apr_pool_t *ptemp) {
  /* Register to handle mod_status status page generation */
  APR_OPTIONAL_HOOK(ap, status_hook, cachem_ext_status_hook, NULL, NULL,
		    APR_HOOK_LAST);
  return OK;
} /* cachem_hook_pre_config*/

static void
cachem_register_hooks(apr_pool_t *p) {
  ap_hook_pre_config(cachem_hook_pre_config, NULL, NULL, APR_HOOK_MIDDLE);
#ifdef DEBUG
  ap_hook_handler(cachem_handler, NULL, NULL, APR_HOOK_MIDDLE);
#endif
  ap_hook_child_init(cachem_hook_child_init, NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_post_config(cachem_hook_post_config, NULL, NULL, APR_HOOK_MIDDLE);
} /* cachem_register_hooks */

command_rec cachem_cmds[]= {
  AP_INIT_TAKE2("CACHEM", cachem_set_shmfile, NULL, RSRC_CONF,
 		"ARG1:Shared memory file name|NULL. ARG2:Cache Size|NULL."),
  AP_INIT_TAKE1("CACHEMlockFile", cachem_set_lockfile, NULL, RSRC_CONF,
		"For AcceptMutex(Lockfile), the path/name of the lockfile"),
  AP_INIT_TAKE1("CACHEMDBM", cachem_set_dbmname, NULL, RSRC_CONF,
		"Absolute path-name of the DBM database"),
  {NULL}
};

/* Dispatch list for API hooks */
module AP_MODULE_DECLARE_DATA cachem_module = {
    STANDARD20_MODULE_STUFF, 
    NULL,                  /* create per-dir    config structures */
    NULL,                  /* merge  per-dir    config structures */
    cachem_per_server_config, /* create per-server config structures */
    NULL,                  /* merge  per-server config structures */
    cachem_cmds,           /* table of config file commands       */
    cachem_register_hooks  /* register hooks                      */
};
