/* 
 *  mod_cachem_private.h -- Apache2 DSO shared memory manager module
 * Byron Young
 * 2005
 */

/**
 *@file mod_cachem_private.h
 *@brief mod_cachem private header file.
 *@author Byron Young
 *@version 0.1
 *@bug report to spamiccling@yahoo.com
 *@bug report to bkyoung@users.sourceforge.net
 */

#ifndef MOD_CACHEM_PRIVATE_H
#define MOD_CACHEM_PRIVATE_H

#ifndef APR_HAS_SHARED_MEMORY
# error Requires shared memroy support
#endif

#ifndef M_INDEX_T
#define M_INDEX_T
typedef unsigned int m_index_t;
#endif /* M_INDEX_T */

#ifndef CACHEM_ID_T
#define CACHEM_ID_T
typedef unsigned int cachem_id_t;
#endif /* CACHEM_ID_T */

/**
 *@struct cachem_info
 *@brief Per process information.
 */
typedef struct cachem_info {
  apr_shm_t * shm; /**< The apr_shm information. */
  void * pData; /**< The start of the data area in the shared memory. */
  apr_size_t nSize; /**< The size of the shared data area. */
  char * dbmname; /**< The id DBM database. */
} cachem_info_t, * p_cachem_info_t;

/**
 *@struct cachem_config
 *@brief Per server (global) configuration information.
 */
typedef struct cachem_config {
  int bSet_shmfile; /**< True if set in config file. */
  int bSet_shmsize; /**< True if set in config file. */
  int bSet_lockfile; /**< True if set in config file. */
  int bSet_dbmname;  /**< True if set in config file. */
  char * sza_shmfile; /**< The named shared memory file name given in config file. */
  apr_size_t shmsize; /**< The size given in configuration file. Must convert to an unsigned integer without loss of data. */
  char * sza_lockfile; /**< The absolute path of the lockfile. */
  char * sza_dbmname; /**< The absolute path of the DBM database. */
} cachem_config_t, * p_cachem_config_t;

#define INITIALIZE_CACHEM_CONFIG {\
 .bSet_shmfile=0, \
 .bSet_shmsize=0, \
 .bSet_lockfile=0, \
 .bSet_dbmname=0, \
 .sza_shmfile=NULL, \
 .shmsize=1024*1, \
 .sza_lockfile=NULL, \
 .sza_dbmname=NULL \
}

/**
 *@struct
 *@brief The first sizeof(cachem_header_t) bytes in the cache.
 */
typedef struct cachem_header {
  unsigned int u_blks; /**< Number of records within cache. */
  unsigned int o_st; /**< The offset of the first byte after the cache_header. */
  unsigned int o_fr; /**< The offset of the first byte after the last record. */
  unsigned int u_size; /**< The size, in bytes, of the cache. */
  unsigned int o_bptop; /**< The offset of the first byte of the last bptr. The top of the block pointers. */
  cachem_id_t u_id; /**< The last unique id released. */
} cachem_header_t, * p_cachem_header_t;

#define CACHEM_DATA_START (APR_ALIGN_DEFAULT(sizeof(cachem_header_t)))

/**
 *@struct cachem_bptr
 *@brief The information about a memory block.
 *The o_beg offset determines the starting offset of the record. It is always
 *__attributes__ ((aligned(4)).
 *The u_size is the size of the record.
 *The actual block size is adjusted to a multiple of 4 ((aligned(4)).
 */
typedef struct cachem_bptr {
  unsigned int o_beg; /**< offset of block */
  unsigned int u_size; /**< total size of record */
  pid_t pid; /**< The pid of the owning process. */
  cachem_id_t id; /**< The cachem id of the module. */
} cachem_bptr_t, * p_cachem_bptr_t;

#define CACHEM_FIRST_BP_OFFSET(pch) (pch->u_size - sizeof(cachem_bptr_t))

/**
 *@struct cachem_cleanup_data
 *brief Cleanup data.
 */
typedef struct cachem_cleanup_data {
  apr_pool_t * p; /**< The pool. */
  int bDetached; /**< True if detached child process. */
  server_rec * s; /**< The server rec. */
} cachem_cleanup_data_t, * p_cachem_cleanup_data_t;

/* PROTOTYPES */
static void cachem_shm_destory(server_rec *s, apr_pool_t *p);
static int cachem_named_shm_init(request_rec *r);
static void cachem_shm_detach();
static void cachem_init();
static m_index_t cachem_next(m_index_t index);
static apr_status_t cachem_post_config_cleanup(void * data);
static int cachem_hook_post_config(apr_pool_t *p, apr_pool_t * plog, apr_pool_t *ptemp, server_rec *s);
static apr_status_t cachem_child_init_cleanup(void * parm);
static void cachem_hook_child_init(apr_pool_t *p, server_rec *s);
static apr_status_t cachem_child_init_cleanup(void * parm);
static const char * cachem_set_shmfile(cmd_parms *parms, void *dummy, const char *arg1, const char * arg2);
static const char * cachem_set_lockfile(cmd_parms *parms, void *dummy, const char *arg);

/* GLOBALS */
static cachem_info_t cachem;
static apr_global_mutex_t * cachem_mutex;

#endif /* MOD_CACHEM_PRIVATE_H */
